from dataclasses import dataclass

from ..apis import OpeniAPI
from ..utils import logger, constants


log = logger.setup_logger()


def get_engine(engine: str):
    try:
        engine_type = constants.API.ENGINE_TYPE.index(engine.lower())
        return engine_type
    except:
        msg = f"❌ `engine` 参数为[pytorch, tensorflow, mindspore, paddlepaddle, oneflow, mxnet, other]其中之一，不区分大小写"
        log.error(msg)
        raise ValueError(msg)


@dataclass
class OpeniModel:
    model_name: str
    repo_id: str
    model_id: str = None
    storage_path: str = None
    model_files: list = None
    model_type: int = None
    model_creator: str = None

    # api = OpeniAPI()

    def __post_init__(self):
        self.api = OpeniAPI()
        self.get_model_id_by_name()
        self.get_model_file()
        log.info(f"OpeniModel Object init: {self.__dict__}")

    def get_model_id_by_name(self):  # repo_id, model_name):
        try:
            response = self.api.model_query_by_name(
                repo_id=self.repo_id,
                model_name=self.model_name,
            ).json()
            self.model_id = response[0]["id"]
            self.storage_path = response[0]["path"]
            self.model_type = response[0]["modelType"]
            self.model_creator = response[0]["userName"]
            self.can_operate = response[0]["isCanOper"]
            self.can_download = response[0]["isCanDownload"]
        except:
            msg = f"❌ `{self.repo_id}` 未找到名为 `{self.model_name}` 的模型"
            log.error(msg)
            raise ValueError(msg)

    def get_model_file(self):
        try:
            response = self.api.model_file_query_by_id(
                repo_id=self.repo_id,
                model_id=self.model_id,
            ).json()
            sorted_model_files = sorted(response, key=lambda x: x["size"])
            self.model_files = sorted_model_files
        except:
            msg = f"❌ `{self.repo_id}`-`{self.model_name}` 获取模型文件列表失败"
            log.error(msg)
            raise ValueError(msg)


def get_model_files_by_id(repo_id, model_id):
    try:
        api = OpeniAPI()
        model_files = api.model_file_query_by_id(
            repo_id=repo_id, model_id=model_id
        ).json()
        return model_files
    except:
        msg = f"❌ "
        log.error(msg)
        raise ValueError(msg)


def create_model(repo_id, model_name):
    api = OpeniAPI()
    response = api.model_create(repo_id, model_name)
    if response.json()["code"] == "0":
        model_id = response.json()["id"]
        return model_id
    else:
        msg = f"model create failed, {response.json()['msg']}"
        log.error(msg)
        raise ValueError(msg)



import os

from ..utils import logger, constants
from ..utils.logger import *
from ..services import *

log = logger.setup_logger()


def download_model(
    repo_id: str,
    model_name: str,
    save_path: str = constants.PATH.MODEL_SAVE_PATH,
):
    """
    download a list of model files from OpenI

    :param repo_id: repo id in form of `owner/repo_name`
    :param model_name: name of model shown on web page
    :param save_path: local save path, default is `./model`
    :return:

    e.g.
    >>> from openi.model import download_model
    >>> download_model("openi/official", "chatglm2b", save_path="mydata")

    Complete(…kenizer_config.json): 100%|█████████████████████████████████████████████████| 244B/244B [00:00<00:00, 795B/s]
    Complete(         config.json): 100%|███████████████████████████████████████████| 1.19kB/1.19kB [00:00<00:00, 1.71kB/s]
    Complete(      .gitattributes): 100%|███████████████████████████████████████████| 1.48kB/1.48kB [00:00<00:00, 1.15kB/s]
    Complete(…guration_chatglm.py): 100%|█████████████████████████████████████████████| 2.19kB/2.19kB [00:01<00:00, 859B/s]
    Complete(       MODEL_LICENSE): 100%|█████████████████████████████████████████████| 4.04kB/4.04kB [00:01<00:00, 686B/s]
    Complete(           README.md): 100%|█████████████████████████████████████████████| 7.89kB/7.89kB [00:01<00:00, 575B/s]
    Complete(…nization_chatglm.py): 100%|█████████████████████████████████████████████| 9.83kB/9.83kB [00:02<00:00, 491B/s]
    Complete(     quantization.py): 100%|█████████████████████████████████████████████| 14.3kB/14.3kB [00:02<00:00, 432B/s]
    Complete(…odel.bin.index.json): 100%|█████████████████████████████████████████████| 20.0kB/20.0kB [00:02<00:00, 384B/s]
    Complete( modeling_chatglm.py): 100%|█████████████████████████████████████████████| 49.5kB/49.5kB [00:02<00:00, 344B/s]
    Complete(     tokenizer.model): 100%|███████████████████████████████████████████████| 995kB/995kB [00:03<00:00, 311B/s]
    Downloading(…-00001-of-00007.bin):   35%|██████████████████████                         | 0.34GB/0.98GB [00:03<?, ?B/s]
    Waiting(…-00002-of-00007.bin):   0%|                                                     | 0.00B/1.69GB [00:00<?, ?B/s]

    """

    openi_repo = OpeniRepo(repo_id=repo_id)
    if openi_repo.access not in ["read", "write"]:
        msg = f"`{repo_id}` 无权限操作此仓库"
        log.error(msg)
        raise ValueError(msg)

    openi_model = OpeniModel(
        model_name=model_name,
        repo_id=openi_repo.repo_id,
    )
    if openi_model.model_files == []:
        msg = f"❌ `{model_name}` 未找到文件"
        log.error(msg)
        raise ValueError(msg)
    if openi_model.can_download is False:
        msg = f"❌ 本机登录用户无此模型下载权限，模型名称 `{openi_model.model_name}`, 仓库名称 `{openi_model.repo_id}`"
        log.error(msg)
        raise ValueError(msg)

    save_path = os.path.expanduser(save_path)
    local_dir = os.path.join(save_path, f"{model_name}")
    if not os.path.exists(local_dir):
        os.makedirs(local_dir, exist_ok=True)

    openifiles = list()
    for targe_file in openi_model.model_files:
        openi_file = OpeniFileDownload(
            filename=targe_file["fileName"],
            size=targe_file["size"],
            local_dir=local_dir,
            repo_id=openi_repo.repo_id,
            dataset_or_model_id=openi_model.model_id,
            dataset_or_model_name=openi_model.model_name,
            target_type="model",
        )
        openifiles.append(openi_file)

    display_progress_bar(openifiles)
    for openi_file in openifiles:
        openi_file.download_with_tqdm()



import argparse

from .login import login, logout, whoami
from .utils.constants import *
from .utils.logger import *
from .dataset import *
from .model import *

# import textwrap


def main():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        usage=CLI.usage,
        description=CLI.banner,
    )
    parser._action_groups.pop()
    subparsers = parser.add_subparsers(title="commands", dest="commands")
    subparsers.required = False
    # subparsers.choices = Help.openi_choices
    parse_login(subparsers)
    parse_logout(subparsers)
    parse_whoami(subparsers)
    parse_dataset(subparsers)
    parse_model(subparsers)
    args = parser.parse_args()
    if not hasattr(args, "func"):
        parser.print_help()
        exit(1)

    command_args = {}
    command_args.update(vars(args))
    del command_args["func"]
    del command_args["commands"]
    error = False
    try:
        out = args.func(**command_args)
    except Exception as e:
        print(e)
        out = None
        error = True
    except ValueError as e:
        print(e)
        out = None
        error = True
    except KeyboardInterrupt:
        print("User cancelled operation")
        out = None
    if out is not None:
        print(out, end="")

    # This is so that scripts that pick up on error codes can tell when there was a failure
    if error:
        exit(1)


def parse_login(subparsers):
    parse_login = subparsers.add_parser(
        "login",
        description=CLI.command_login,
        usage=CLI.login_usage,
        help=CLI.command_login,
    )
    parse_login.add_argument(
        "-e",  # '--endpoint',
        dest="endpoint",
        default=API.DEFAULT_ENDPOINT,
        required=False,
        help=CLI.param_endpoint,
    )
    parse_login.add_argument(
        "-t",  # '--token',
        dest="token",
        default=None,
        required=False,
        help=CLI.param_token,
    )
    parse_login.set_defaults(func=login)


def parse_logout(subparsers):
    parse_logout = subparsers.add_parser(
        "logout",
        description=CLI.command_logout,
        usage="openi logout [-h]",
        help=CLI.command_logout,
    )
    parse_logout.set_defaults(func=logout)


def parse_whoami(subparsers):
    parse_whoami = subparsers.add_parser(
        "whoami",
        description=CLI.command_whoami,
        usage="openi whoami [-h]",
        help=CLI.command_whoami,
    )
    parse_whoami.set_defaults(func=whoami)


def parse_dataset(subparsers):
    parser_dataset = subparsers.add_parser(
        "dataset", usage=CLI.dataset_usage, help=CLI.command_dataset, aliases=["d"]
    )
    parser_dataset._action_groups.pop()
    subparsers_dataset = parser_dataset.add_subparsers(
        title="commands", dest="commands"
    )
    subparsers_dataset.required = True

    # dataset upload
    parser_dataset_upload = subparsers_dataset.add_parser(
        "upload",
        description=CLI.command_dataset_upload,
        usage=CLI.dataset_upload_usage,
        help=CLI.dataset_upload_help,
        # epilog=CLI.dataset_upload_epilog,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser_dataset_upload._action_groups.pop()
    parser_dataset_upload_args = parser_dataset_upload.add_argument_group("arguments")
    parser_dataset_upload_args.add_argument(
        "-f", dest="file", required=True, help=CLI.dataset_upload_param_file
    )
    parser_dataset_upload_args.add_argument(
        "-r", dest="repo_id", required=True, help=CLI.dataset_upload_param_repo_id
    )
    parser_dataset_upload_args.add_argument(
        "-c",
        dest="cluster",
        required=False,
        default="NPU",
        choices=["gpu", "npu"],
        help=CLI.dataset_upload_param_cluster,
    )
    parser_dataset_upload.set_defaults(func=upload_file)

    # dataset download
    parser_dataset_download = subparsers_dataset.add_parser(
        "download",
        description=CLI.command_dataset_download,
        usage=CLI.dataset_download_usage,
        help=CLI.dataset_download_help,
        # epilog = CLI.dataset_download_epilog,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser_dataset_download._action_groups.pop()
    parser_dataset_download_args = parser_dataset_download.add_argument_group(
        "arguments"
    )
    parser_dataset_download_args.add_argument(
        "-f", dest="file", required=True, help=CLI.dataset_download_param_file
    )
    parser_dataset_download_args.add_argument(
        "-r", dest="repo_id", required=True, help=CLI.dataset_upload_param_repo_id
    )
    parser_dataset_download_args.add_argument(
        "-c",  #'--cluster',
        dest="cluster",
        required=False,
        default="npu",
        choices=["gpu", "npu"],
        help=CLI.dataset_upload_param_cluster,
    )
    parser_dataset_download_args.add_argument(
        "-p",  #'--save_path',
        dest="save_path",
        required=False,
        default=PATH.DATASET_SAVE_PATH,
        help=CLI.dataset_download_param_save_path,
    )
    parser_dataset_download.set_defaults(func=download_file)


def parse_model(subparsers):
    parser_model = subparsers.add_parser(
        "model", usage=CLI.model_usage, help=CLI.command_model, aliases=["m"]
    )
    parser_model._action_groups.pop()
    subparsers_model = parser_model.add_subparsers(title="commands", dest="commands")
    subparsers_model.required = True

    # model upload
    parser_model_upload = subparsers_model.add_parser(
        "upload",
        description=CLI.command_model_upload,
        usage=CLI.model_upload_usage,
        help=CLI.model_upload_help,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser_model_upload._action_groups.pop()
    parser_model_upload_args = parser_model_upload.add_argument_group("arguments")
    parser_model_upload_args.add_argument(
        "-f", dest="folder", required=True, help=CLI.model_upload_param_folder
    )
    parser_model_upload_args.add_argument(
        "-r", dest="repo_id", required=True, help=CLI.model_upload_param_repo_id
    )
    parser_model_upload_args.add_argument(
        "-m",
        dest="model_name",
        required=True,
        help=CLI.model_upload_param_model_name,
    )
    parser_model_upload.set_defaults(func=upload_model)

    # model upload
    parser_model_download = subparsers_model.add_parser(
        "download",
        description=CLI.command_model_download,
        usage=CLI.model_download_usage,
        help=CLI.model_download_help,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser_model_download._action_groups.pop()
    parser_model_download_args = parser_model_download.add_argument_group("arguments")
    parser_model_download_args.add_argument(
        "-r", dest="repo_id", required=True, help=CLI.model_upload_param_repo_id
    )
    parser_model_download_args.add_argument(
        "-m",
        dest="model_name",
        required=True,
        help=CLI.model_upload_param_model_name,
    )
    parser_model_download_args.add_argument(
        "-p",  #'--save_path',
        dest="save_path",
        required=False,
        default=PATH.MODEL_SAVE_PATH,
        help=CLI.model_download_param_save_path,
    )
    parser_model_download.set_defaults(func=download_model)



import hashlib
import math
import os
from dataclasses import dataclass

from .file_progress_bar import *
from ..apis import OpeniAPI
from ..utils import logger, constants
from ..utils.logger import *

log = logger.setup_logger()


@dataclass
class OpeniFileUpload:
    """
    Upload file object
    """

    local_file: str
    repo_id: str
    upload_type: str
    dataset_or_model_id: str
    dataset_or_model_name: str
    target_type: str  # ['dataset','model']
    chunk_size: int = None
    total_chunks_count: int = None
    pbar: object = None
    md5_pbar: object = None

    uploaded: int = False  # 1 for True, 0 for False
    uploading_chunks_idx: list = None
    uuid: str = None
    upload_id: str = None
    filename_exists: str = None

    def __post_init__(self):
        self.api = OpeniAPI()

        self.filename = os.path.basename(self.local_file)
        self.filepath = os.path.abspath(self.local_file)

        self.size = os.path.getsize(self.filepath)
        if self.md5_pbar is not None:
            self.md5 = calculateMD5(self.filepath, self.md5_pbar)
        else:
            self.md5 = calculateMD5(self.filepath)
        self.set_chunk_size()
        self.set_total_chunks_count()

        log.info(f"OpeniFileUpload Object init: {self.__dict__}")

    def set_chunk_size(self):
        if self.target_type == "model":
            self.chunk_size = constants.FILE.LARGE_FILE_CHUNK_SIZE
        else:
            if self.size < constants.FILE.SMALL_FILE_LIMIT:
                self.chunk_size = constants.FILE.SMALL_FILE_CHUNK_SIZE
            else:
                self.chunk_size = constants.FILE.LARGE_FILE_CHUNK_SIZE

    def set_total_chunks_count(self):
        total_chunks_count = math.ceil(self.size / self.chunk_size)
        self.total_chunks_count = total_chunks_count

    def read_file_by_chunk(self, real_chunk_size: int, start_position: int = 0):
        with open(self.filepath, "rb") as file:
            file.seek(start_position)
            chunk = file.read(real_chunk_size)
            return chunk

    def get_chunks(self):
        if self.target_type == "dataset":
            get_chunks = self.api.dataset_file_get_chunks(
                dataset_id=self.dataset_or_model_id,
                md5=self.md5,
                filename=self.filename,
                upload_type=self.upload_type,
            ).json()
            if "repoName" in get_chunks and "repoOwner" in get_chunks:
                cluster = constants.FILE.UPLOAD_TYPE[self.upload_type]
                self.pbar.close()
                msg = (
                    f"Falied: ❌ `{get_chunks['fileName']}` ({cluster})"
                    " 你已上传过此数据集，无法重复上传，点击链接查看 "
                    f"{self.api.endpoint.split('/api')[0]}/{get_chunks['repoOwner']}/{get_chunks['repoName']}/datasets"
                )
                log.error(msg)
                raise ValueError(msg)
        if self.target_type == "model":
            get_chunks = self.api.model_file_get_chunks(
                modeluuid=self.dataset_or_model_id,
                md5=self.md5,
                filename=self.filename,
                upload_type=self.upload_type,
            ).json()

        # upload completed
        self.uploaded = get_chunks["uploaded"]

        # continue upload
        if get_chunks["uuid"] != "" and get_chunks["uploadID"] != "":
            self.uuid = get_chunks["uuid"]
            self.upload_id = get_chunks["uploadID"]
            uploaded_chunks = [i for i in get_chunks["chunks"].split(",") if i != ""]
            # init uuid and uploadID but no chunks
            init_idx = 1 if uploaded_chunks == [] else len(uploaded_chunks)
            self.uploading_chunks_idx = [
                i for i in range(init_idx, self.total_chunks_count + 1)
            ]

    def new_multipart(self):
        if self.target_type == "dataset":
            new_multipart = self.api.dataset_file_new_multipart(
                dataset_id=self.dataset_or_model_id,
                md5=self.md5,
                filename=self.filename,
                upload_type=self.upload_type,
                size=self.size,
                total_chunks_count=self.total_chunks_count,
            ).json()
        if self.target_type == "model":
            new_multipart = self.api.model_file_new_multipart(
                modeluuid=self.dataset_or_model_id,
                md5=self.md5,
                filename=self.filename,
                upload_type=self.upload_type,
                size=self.size,
                total_chunks_count=self.total_chunks_count,
            ).json()

        self.uuid = new_multipart["uuid"]
        self.upload_id = new_multipart["uploadID"]
        self.uploading_chunks_idx = [i for i in range(1, self.total_chunks_count + 1)]

    def get_multipart_url(self, chunk_number: int):
        if self.target_type == "dataset":
            url = self.api.dataset_file_get_multipart_url(
                chunk_number=chunk_number,
                chunk_size=self.chunk_size,
                dataset_id=self.dataset_or_model_id,
                filename=self.filename,
                upload_type=self.upload_type,
                upload_id=self.upload_id,
                uuid=self.uuid,
            ).json()["url"]
        if self.target_type == "model":
            url = self.api.model_file_get_multipart_url(
                chunk_number=chunk_number,
                chunk_size=self.chunk_size,
                upload_type=self.upload_type,
                upload_id=self.upload_id,
                uuid=self.uuid,
            ).json()["url"]

        return url

    def complete_multipart(self):
        if self.target_type == "dataset":
            self.api.dataset_file_complete_multipart(
                dataset_id=self.dataset_or_model_id,
                filename=self.filename,
                upload_type=self.upload_type,
                size=self.size,
                upload_id=self.upload_id,
                uuid=self.uuid,
            )
        if self.target_type == "model":
            self.api.model_file_complete_multipart(
                modeluuid=self.dataset_or_model_id,
                upload_type=self.upload_type,
                upload_id=self.upload_id,
                uuid=self.uuid,
            )

    def upload_with_tqdm(self):
        # get_chunks
        self.get_chunks()

        # if uploaded and complete: update bar, return
        if self.uploaded == "1":
            msg = f"Failed: ❌ 已存在, 无需重复上传"
            self.pbar.set_description_str(self.pbar.desc.replace("Waiting", msg))
            # self.pbar.close()
            log.error(self.pbar.desc)
            return

        if self.uploaded == "0":
            # if continue: get_multipart_url, update bar, upload_chunks, complete_multipart
            # if not complete: update bar, complete_multipart
            # if new: new_multipart, update bar, upload_chunks, complete_multipart
            if not self.uuid and not self.upload_id:
                self.new_multipart()

            # start uploading
            self.pbar.set_description_str(
                self.pbar.desc.replace("Waiting", "Uploading")
            )
            init_progress = self.chunk_size * (
                self.total_chunks_count - len(self.uploading_chunks_idx)
            )
            self.pbar.update(init_progress)

            # upload_chunks
            for n in self.uploading_chunks_idx:
                start_position = (n - 1) * self.chunk_size
                real_chunk_size = min(self.chunk_size, self.size - start_position)

                data = self.read_file_by_chunk(
                    real_chunk_size=real_chunk_size, start_position=start_position
                )
                for i in range(real_chunk_size):
                    self.pbar.update(1)

                url = self.get_multipart_url(chunk_number=n)
                response = self.api.file_put_upload(
                    url=url,
                    data=data,
                    upload_type=self.upload_type,
                )
                etag = response.headers.get("ETag")
                if etag is None:
                    msg = f"❌ file_put_upload failed: {self.filename} chunk {n} failed to upload."
                    log.error(msg)
                    raise RuntimeError(msg)

            # complete upload process
            self.complete_multipart()
            self.pbar.set_description_str(
                self.pbar.desc.replace("Uploading", "Complete")
            )
            # self.pbar.close()


"""
public methods
"""


def calculateMD5(filepath: str, md5_pbar: object = None) -> str:
    """
    Calculate MD5 of a file
    """
    m = hashlib.md5()
    with open(filepath, "rb") as f:
        while True:
            data = f.read(4096)
            if not data:
                break
            m.update(data)
    return m.hexdigest()


def get_upload_folder_files(folder: str):
    local_dir = os.path.expanduser(folder)
    files = list()
    for root, dirs, filenames in os.walk(local_dir):
        if dirs != []:
            msg = f"❌ {local_dir} 路径下不允许有子文件夹."
            log.error(msg)
            raise ValueError(msg)
        for file in filenames:
            filepath = os.path.join(root, file)
            filesize = os.path.getsize(filepath)
            upload_size_check(file, filesize)
            files.append([filepath, filesize])
    sorted_files = [i[0] for i in sorted(files, key=lambda x: x[1])]
    return local_dir, sorted_files


def upload_size_check(filename: str, size: int):
    # msg = None
    if size == 0:
        msg = f"❌ `{filename}` file size is 0"
        log.error(msg)
        raise ValueError(msg)
    if size > constants.FILE.MAX_FILE_SIZE:
        msg = f"❌ `{filename}` file size exceeds {constants.FILE.MAX_FILE_SIZE_GB}GB"
        log.error(msg)
        raise ValueError(msg)
